#include "header.hpp"
#include <array>

#include "ppaaa23.amplitude.doublevirtual.hpp"

#include "FivePointAmplitudes/ME2/qqyyy_h2.h"


/**
 * Select which contributions to include in H2.
 * According to definitions in the paper in this order:
 * {H(2,0), H(2,1), H(2,Nf), H(2,Nfy)}
 * set to true to include. 
 *
 * TODO: make this to be not hard coded
 */
static thread_local std::array<bool,4> ppaaa23_H2_included_twoloop_contributions = {true, false, false, false};

void ppaaa23_calculate_H2(observable_set* osi) {
    static Logger logger("ppaaa23_calculate_H2");
    logger << LOG_DEBUG_VERBOSE << "called" << endl;

    static int initialization = 1;
    if (initialization){
      if (osi->user->switch_value[osi->user->switch_map["switch_on_Nf"]] == 1){
	ppaaa23_H2_included_twoloop_contributions[0] = 0;
	ppaaa23_H2_included_twoloop_contributions[2] = 1;
      }
      for (int i = 0; i < 4; i++){
	logger << LOG_INFO << "ppaaa23_H2_included_twoloop_contributions[" << i << "] = " << ppaaa23_H2_included_twoloop_contributions[i] << endl;
      }
      initialization = 0;
    }
    
    using R = double;
    using std::pow;
    //dbg::g_floating_point_format << std::scientific << std::setprecision(17);

    /**
     * NOTE:
     * The functions hi_catani are normalized by the born ME2 such that the leading pole 
     * of the loop ME2 before IR-subtractions is given exactly by the I_i operators.
     * ALL factors (leading-order couplings, helicity averaging, etc.) will be taken from internal born ME2 in MATRIX.
     * hi_catani[0] is the born ME2 up to *some* constant prefactor, which we do not need to know here.
     * Additionally, color factors are *not* included in these functions, we restore them below.
     *
     * Made static such that the construction stage is not repeated.
     */
    static thread_local FivePointAmplitudes::qqyyy_h2<R> hi_evaluator{ppaaa23_H2_included_twoloop_contributions};

    const auto& mm = osi->esi->p_parton[0];

    std::array<R, 5> sij = {
        (mm[1] + mm[2]).m2(),
        (mm[2] - mm[3]).m2(),
        (mm[3] + mm[4]).m2(),
        (mm[4] + mm[5]).m2(),
        (mm[1] - mm[5]).m2(),
    };

    auto hi_catani = hi_evaluator(sij, sij[0]);

    R half_N_c = physconst::N_c / R(2);

    // first restore (leading) color factors for results in Catani subtraction scheme
    hi_catani[0][0] *= physconst::N_c;
    hi_catani[1][0] *= half_N_c; // at full color this factor is Cf
    // 2 from 4 two-loop contributions are included so far
    hi_catani[2][0] *= pow(half_N_c, 2);
    hi_catani[2][2] *= ((physconst::C_F* osi->N_f)/2); // here division by half is from T_F = 1/2

    std::array<R,3> hi_qt = {hi_catani[0][0],hi_catani[1][0],R(0)};

    // finite shift of the hard functions from Catani to qT scheme

    hi_qt[1] += half_N_c * mathconst::pi2;

    // shift proportional to Nc^2/4
    if (ppaaa23_H2_included_twoloop_contributions[0]) {
        hi_qt[2] += hi_catani[2][0];
        // WARNING:
        // this shift will be also related to the contribution from subleading color, since hi_catani[1][0] will by multiplied by Cf at full color
        hi_qt[2] += half_N_c * mathconst::pi2 * hi_catani[1][0];

        hi_qt[2] += pow(half_N_c, 2) *
                    (R(1214) / 81 + (1181 * mathconst::pi2) / 108 +
                     (5 * mathconst::pi4) / 24 - (187 * mathconst::zeta3) / 18);
    }
    // shift proportional to Cf*Nf*Tf
    if(ppaaa23_H2_included_twoloop_contributions[2]) {
        hi_qt[2] += hi_catani[2][2];
        hi_qt[2] += ((physconst::C_F* osi->N_f)/2) * (
                -(R(164)/81) - (97*mathconst::pi2)/54 + (17*mathconst::zeta3)/9
        );
    }

    // in this case leading-color approximation for H1 can be converted to the full-color result by simple rescaling
    hi_qt[1] *= physconst::C_F/half_N_c;

    osi->QT_A0 = osi->VA_b_ME2;
    osi->QT_A1 = pow(osi->alpha_S * mathconst::inv2pi, 1) * (osi->QT_A0) * hi_qt[1];
    osi->QT_A2 = pow(osi->alpha_S * mathconst::inv2pi, 2) * (osi->QT_A0) * hi_qt[2];

    // additional factors from (alpha_S/2*Pi) -> (alpha_S/Pi) definition of perturbative expansion of H2
    hi_qt[1] /= 2;
    hi_qt[2] /= 4;


    //dbg((osi->VA_V_ME2) / 2 / (osi->alpha_S * mathconst::inv2pi * osi->VA_b_ME2)); // H1 from OpenLoops
    //dbg(hi_qt);

    osi->QT_H1_delta_QCD = hi_qt[1];  // full color
    osi->QT_H2_delta_QCD = hi_qt[2];  // leading color

    logger << LOG_DEBUG_VERBOSE << "finished" << endl;
}

