#include "header.hpp"

void xdistribution::computeObservable(double &observable, double &observable_max) {
  static Logger logger("xdistribution::computeObservable");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  logger << LOG_DEBUG_VERBOSE << xdistribution_name << endline_flush;

  observable = 0.;
  observable_max = 0.;

  // To be added somwhere:
  // 'reconstructedParticle[i-1] == particle i' in distribution.dat
  // .eta() is class function to compute pseudo-rapidity

  ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //
  //////////////////////////////////////////////////////////
  //  Short Tuturial for creating your own distributions  //
  //////////////////////////////////////////////////////////
  //
  // There can be several groups 'particle i' (i=1,2,3,...) defined in distribution.dat
  // The behavior when several particle groups are defined depends on the observable
  //
  // There can also be several particles in one group, eg:
  // particle 1 = lep 1  # hardest lepton
  // particle 1 = lep 2  # second-hardest lepton
  // The momentum of group 'particle 1' is then computed as vectorial sum of the 4-momenta:
  // p(particle 1) = p(lep 1) + p(lep 2)
  //
  // The reconstructed (vectorial) sum of each particle group is accessible below by:
  // reconstructedParticles[group]
  // where
  // reconstructedParticles[group] == 'particle group+1' in distribution.dat, ie:
  // reconstructedParticles[0] == 'particle 1'
  // reconstructedParticles[1] == 'particle 2'
  // reconstructedParticles[2] == 'particle 3'
  // ...
  //
  //
  // 4-momenta (fourvector class)
  // ----------------------------
  //
  // The C++ type of 'reconstructedParticles' is 'fourvector', ie a 4-momentum.
  // 4-momenta can be summed (use 'fourvector myfourvector' to define a new fourvector), eg:
  // fourvector sum
  // fourvector electron = reconstructedParticles[0]
  // fourvector muon     = reconstructedParticles[1]
  // sum = electron + muon
  //
  // or direclty:
  // fourvector sum
  // sum = reconstructedParticles[0] + reconstructedParticles[1]
  //
  //
  // Define observables
  // ------------------
  //
  // Some observables are already predefined through functions in the class and can be directly accessed, eg:
  // reconstructedParticles[0].pT()   gives the transverse momentum pT
  // reconstructedParticles[0].eta()  gives the pseudo-rapidity eta
  // ...
  // see the class definition (src-MUNICH/classes/fourvector.cpp) and header (src-MUNICH/classes/header/fourvector.h)
  // for more further pre-defined observables.
  //
  // ALL (other) observables can be computed directly from the 4-momenta:
  // E = reconstructedParticles[0].x0()
  // x = reconstructedParticles[0].x1()
  // y = reconstructedParticles[0].x2()
  // z = reconstructedParticles[0].x3()
  //
  //
  //
  // Define new distribution
  // -----------------------
  //
  // Follow the pre-defined distributions below and define your own 'xdistribution_type'. Copy a an 'else if(...){...}'
  // block, choose a unique name of your distribition and define your 'observable', eg:
  //
  // ...
  // else if (xdistribution_type == "unique_distribution_name") {
  //   assert(particles.size()==1 && "ERROR in xdistribution_type unique_distribution_name: only one particle group allowed");
  //   E = reconstructedParticles[0].x0()
  //   x = reconstructedParticles[0].x1()
  //   y = reconstructedParticles[0].x2()
  //   z = reconstructedParticles[0].x3()
  //   observable = E+x+y+z
  // }
  // ...
  //
  ////////////////////////////////////////////////////////////////////////////////////////////////////////////////////



  if (xdistribution_type == "XS"){
    observable = 0.;
  }
  else if (xdistribution_type == "multiplicity"){
    observable = particles[0][0].multiplicity;
  }
  else if (xdistribution_type == "multiplicity+"){
    observable_max = particles[0][0].multiplicity;
    // FIXME: more general version not ported. still needed?
  }
  else if (xdistribution_type == "muR"){
    // bins at the renormalization scale
    assert(particles.size()==0);
    observable = osi->var_mu_ren;
  }
  else if (xdistribution_type == "muF"){
    // bins at the factorization scale
    assert(particles.size()==0);
    observable = osi->var_mu_fact;
  }

  ///////////////////////////////////////////////////////////////////
  //  cumulative taucut distribution of particle/sum of particles  //
  ///////////////////////////////////////////////////////////////////

  else if (xdistribution_type == "taucut"){
    // temporary N-jettiness implementation (never applied so far, I guess !!!)
    static int Njettiness = 0;
    static vector<fourvector> NJ_axes;
    Njettiness = osi->csi->n_jet_born;
    //    Njettiness = osi->order_alphas_born;
    NJ_axes.resize(3 + Njettiness); // not generic !!! there should be n_jet_born or so... !!!
    NJ_axes[1] = fourvector(1., 0., 0., 1.);
    NJ_axes[2] = fourvector(1., 0., 0., -1.);
    double tau_jettiness = 0.;
    for (int i_j = 0; i_j < Njettiness; i_j++){
      //      fourvector temp_jet = osi->esi->event_jet[0][osi->RA_x_a][i_j].momentum;
      fourvector temp_jet = osi->esi->event_particle_group[0][osi->RA_x_a][osi->esi->access_object_selection["jet"]][i_j]->momentum;
      double temp_jet_R = temp_jet.r();
      NJ_axes[3 + i_j] = fourvector(1., temp_jet.x1() / temp_jet_R, temp_jet.x2() / temp_jet_R, temp_jet.x3() / temp_jet_R);
    }
    // all available QCD partons
    stringstream temp;
    logger << LOG_DEBUG_VERBOSE << "osi->esi->index_group_protojet_parton[osi->RA_x_a].size() = " << osi->esi->index_group_protojet_parton[osi->RA_x_a].size() << endline_flush;

    for (int i_p = 0; i_p < osi->esi->index_group_protojet_parton[osi->RA_x_a].size(); i_p++){temp << setw(5) << osi->esi->index_group_protojet_parton[osi->RA_x_a][i_p];}
    logger << LOG_DEBUG_VERBOSE << "Njettiness = " << Njettiness << "   osi->esi->index_group_protojet_parton[" << osi->RA_x_a << "].size() = " << osi->esi->index_group_protojet_parton[osi->RA_x_a].size() << " --- " << temp.str() << endline_flush;
    for (int i_j = 1; i_j < NJ_axes.size(); i_j++){
      logger << LOG_DEBUG_VERBOSE << "NJ_axes[" << i_j << "] = " << NJ_axes[i_j] << endline_flush;
    }
    // calculate N-jettiness tau
    for (int i_p = 0; i_p < osi->esi->index_group_protojet_parton[osi->RA_x_a].size(); i_p++){
      vector<double> delta_tau(3 + Njettiness);
      for (int i_j = 1; i_j < NJ_axes.size(); i_j++){
	delta_tau[i_j] = abs(NJ_axes[i_j] * osi->esi->event_particle_group[0][osi->RA_x_a][0][osi->esi->index_group_protojet_parton[osi->RA_x_a][i_p]]->momentum);
	//	  delta_tau[i_j] = NJ_axes[i_j] * osi->p_parton[osi->RA_x_a][osi->esi->index_group_protojet_parton[osi->RA_x_a][i_p]];
	logger << LOG_DEBUG_VERBOSE << "i_p = " << i_p << "   delta_tau[" << i_j << "] = " << delta_tau[i_j] << endline_flush;

	if (delta_tau[i_j] < 0.){
	  //	  logger << LOG_DEBUG_VERBOSE << "NJ_axes[" << i_j << "] * event_particle_group[0][" << osi->RA_x_a << "][0][" << osi->esi->index_group_protojet_parton[osi->RA_x_a][i_p] << "] is negative: delta_tau[" << i_j << "] = " << delta_tau[i_j] << endline_flush;

	  delta_tau[i_j] = abs(delta_tau[i_j]);
	}
      }

      double min = 1.e99;
      for (int i_j = 1; i_j < NJ_axes.size(); i_j++){
	if (delta_tau[i_j] < min){min = delta_tau[i_j];}
      }
      //      logger << LOG_DEBUG_VERBOSE << "min = " << min << " =?= " << *min_element(delta_tau.begin(), delta_tau.end()) << " = *min_element(delta_tau.begin(), delta_tau.end())" << endline_flush;
      logger << LOG_DEBUG_VERBOSE << "i_p = " << i_p << "   tau_jettiness += " << *min_element(delta_tau.begin(), delta_tau.end()) << endline_flush;
      tau_jettiness += *min_element(delta_tau.begin() + 1, delta_tau.end());
      //	tau_jettiness = tau_jettiness + *min_element(delta_tau.begin(), delta_tau.end());
    }
    logger << LOG_DEBUG_VERBOSE << "tau_jettiness = " << tau_jettiness << endline_flush;
    observable_max += tau_jettiness;
  }



  /////////////////////////////////////////
  //  transverse-momentum distributions  //
  /////////////////////////////////////////


  ///////////////////////////////////////////////////////////////////////////////////////////
  //  transverse-momentum distribution of particle group (scalar sum when several groups)  //
  //  pT = sum_i pT(particle i);  pT(particle i) = sqrt(p_x1^2 + p_x2^2)                   //
  ///////////////////////////////////////////////////////////////////////////////////////////
  else if (xdistribution_type == "pT") {
    observable = 0.;
    // loop over all particle groups (group=0,1,2,..., 'particle group+1' distribution.dat)
    for (int group=0; group<particles.size(); group++) {
      observable += particle_pointer[group]->pTXX(); // scalar sum of pT of all particle groups
    }
  }
  ///////////////////////////////////////
  //  pTmax distribution of particles  //
  ///////////////////////////////////////
  else if (xdistribution_type == "pTmax") {
    double max_pT=0;
    for (int group=0; group<particles.size(); group++) {
      double tmp_pT=particle_pointer[group]->pTXX();
      if (tmp_pT>max_pT) {
        max_pT=tmp_pT;
      }
    }
    observable = max_pT;
  }
  ///////////////////////////////////////
  //  pTmin distribution of particles  //
  ///////////////////////////////////////
  else if (xdistribution_type == "pTmin") {
    double min_pT=1e99;
    for (int group=0; group<particles.size(); group++) {
      double tmp_pT=particle_pointer[group]->pTXX();
      if (tmp_pT<min_pT) {
        min_pT=tmp_pT;
      }
    }
    observable = min_pT;
  }
  ////////////////////////////////////////////////////
  //  pT ratio of particle/sum of particles         //
  ////////////////////////////////////////////////////
  else if (xdistribution_type == "pTratio") {
    assert(particles.size()==2);

    observable = particle_pointer[0]->pTXX() / particle_pointer[1]->pTXX();
    if (abs(observable - 1.) < 1.e-12){observable = 1. - 1.e-12;} // in case of back-to-back particles the ratio should be exactly one !!!
  }
  ////////////////////////////////////////////////////
  //  pT-over-m ratio of particle/sum of particles  //
  ////////////////////////////////////////////////////
  else if (xdistribution_type == "pTbyM") {
    assert(particles.size()==1);

    observable = particle_pointer[0]->pTXX() / particle_pointer[0]->mXX();
    //    if (abs(observable - 1.) < 1.e-12){observable = 1. - 1.e-12;} // in case of back-to-back particles the ratio should be exactly one !!!
  }
  //////////////////////////////////////
  //  projected ET_miss distribution  //
  //////////////////////////////////////
  else if (xdistribution_type == "ETproj") {
    assert(particles.size()>1);

    // pT from neutrinos is the first reconstructed particle
    double phi_missing = particle_pointer[0]->phiXX();

    // now find minimum delta phi between this and the other particles
    double dphi_min=0.5*pi;
    for (int group=1; group<particles.size(); group++) {
      double dphi_tmp = min(f2pi-abs(phi_missing-particle_pointer[group]->phiXX()),abs(phi_missing-particle_pointer[group]->phiXX()));
      if (dphi_tmp<dphi_min) {
        dphi_min = dphi_tmp;
      }
    }
    observable = particle_pointer[0]->pTXX()*sin(dphi_min);
  }
  //////////////////////////////////////////////////////////
  //  pTveto 'distribution' of particle/sum of particles  //
  //////////////////////////////////////////////////////////
  else if (xdistribution_type == "pTveto"){
    assert(particles.size()==1);
    observable = particle_pointer[0]->pTXX();
  }
  //////////////////////////////////////////////////////////
  //  'exclusive' pTveto 'distribution' of particle/sum of particles  //
  //////////////////////////////////////////////////////////
  else if (xdistribution_type == "pTvetoexcl"){
    assert(particles.size()==2);
    observable = particle_pointer[0]->pTXX();
    observable_max = particle_pointer[1]->pTXX();
  }
  //////////////////////////////////////////////////////////
  //  pT-cut 'distribution' of particle/sum of particles  //
  //////////////////////////////////////////////////////////
  else if (xdistribution_type == "pTcut"){
    for (int group=0; group<particles.size(); group++) {
      observable_max += particle_pointer[group]->pTXX();
    }
  }



  //  new cumulative distributiona for HE-LHC and FCC benchmarks

  /////////////////////////////////////////////////////////////
  //  pTmin-cut 'distribution' of particle/sum of particles  //
  /////////////////////////////////////////////////////////////
  else if (xdistribution_type == "pTmincut"){
    observable_max = 1.e99;
    for (int group = 0; group < particles.size(); group++){
      if (observable_max > particle_pointer[group]->pTXX()){observable_max = particle_pointer[group]->pTXX();}
    }
  }

  ////////////////////////////////////////////////////////////////
  //  maximal |pseudo-rapidity| veto distribution of particle:  //
  ////////////////////////////////////////////////////////////////
  else if (xdistribution_type == "absetamaxveto"){
    //    assert(particles.size() == 1 && "ERROR in xdistribution_type 'abseta': exactly one particle required");
    observable = 0.;
    for (int group = 0; group < particles.size(); group++){
      if (observable < abs(particle_pointer[group]->etaXX())){observable = abs(particle_pointer[group]->etaXX());}
    }
  }

  //////////////////////////////////////////////////////////////////////////////////////////////////
  //  maximal |pseudo-rapidity| veto distribution of particle (extra condition: pTlep > 20 GeV):  //
  //////////////////////////////////////////////////////////////////////////////////////////////////
  else if (xdistribution_type == "absetamaxvetopTlep20"){
    //    assert(particles.size() == 1 && "ERROR in xdistribution_type 'abseta': exactly one particle required");
    int cut_condition = 0;
    if (osi->esi->event_particle_group[0][osi->RA_x_a][osi->esi->access_object_selection["nua"]].size() > 0){
      if (osi->esi->event_particle_group[0][osi->RA_x_a][osi->esi->access_object_selection["missing"]][0]->pTXX() < 20.){cut_condition = 1;}
    }

    if (!cut_condition){
      for (int i_p = 0; i_p < osi->esi->event_particle_group[0][osi->RA_x_a][osi->esi->access_object_selection["lep"]].size(); i_p++){
	if (osi->esi->event_particle_group[0][osi->RA_x_a][osi->esi->access_object_selection["lep"]][i_p]->pTXX() < 20.){cut_condition = 1; break;}
      }
    }

    if (cut_condition){
      observable = 1.e99;
    }
    else {
      observable = 0.;
      for (int group = 0; group < particles.size(); group++){
	if (observable < abs(particle_pointer[group]->etaXX())){observable = abs(particle_pointer[group]->etaXX());}
      }
    }
  }

  ///////////////////////////////////////////////////////////////////////////////////////////////////
  //  maximal |pseudo-rapidity| veto distribution of particle (extra condition: pTlep > 100 GeV):  //
  ///////////////////////////////////////////////////////////////////////////////////////////////////
  else if (xdistribution_type == "absetamaxvetopTlep100"){
    //    assert(particles.size() == 1 && "ERROR in xdistribution_type 'abseta': exactly one particle required");
    int cut_condition = 0;
    if (osi->esi->event_particle_group[0][osi->RA_x_a][osi->esi->access_object_selection["nua"]].size() > 0){
      if (osi->esi->event_particle_group[0][osi->RA_x_a][osi->esi->access_object_selection["missing"]][0]->pTXX() < 100.){cut_condition = 1;}
    }
    if (!cut_condition){
      for (int i_p = 0; i_p < osi->esi->event_particle_group[0][osi->RA_x_a][osi->esi->access_object_selection["lep"]].size(); i_p++){
	if (osi->esi->event_particle_group[0][osi->RA_x_a][osi->esi->access_object_selection["lep"]][i_p]->pTXX() < 100.){cut_condition = 1; break;}
      }
    }

    if (cut_condition){
      observable = 1.e99;
    }
    else {
      observable = 0.;
      for (int group = 0; group < particles.size(); group++){
	if (observable < abs(particle_pointer[group]->etaXX())){observable = abs(particle_pointer[group]->etaXX());}
      }
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  //  cumulative invariant-mass 'distribution' of particle/sum of particles  //
  /////////////////////////////////////////////////////////////////////////////
  else if (xdistribution_type == "mcutpTlep20" || xdistribution_type == "McutpTlep20"){
    int cut_condition = 0;
    if (osi->esi->event_particle_group[0][osi->RA_x_a][osi->esi->access_object_selection["nua"]].size() > 0){
      if (osi->esi->event_particle_group[0][osi->RA_x_a][osi->esi->access_object_selection["missing"]][0]->pTXX() < 20.){cut_condition = 1;}
    }
    if (!cut_condition){
      for (int i_p = 0; i_p < osi->esi->event_particle_group[0][osi->RA_x_a][osi->esi->access_object_selection["lep"]].size(); i_p++){
	if (osi->esi->event_particle_group[0][osi->RA_x_a][osi->esi->access_object_selection["lep"]][i_p]->pTXX() < 20.){cut_condition = 1; break;}
      }
    }

    if (cut_condition){
      observable_max = -1.;
    }
    else {
      for (int i_g = 0; i_g < particles.size(); i_g++) {
	observable_max += particle_pointer[i_g]->mXX();
      }
    }
  }

  /////////////////////////////////////////////////////////////////////////////
  //  cumulative invariant-mass 'distribution' of particle/sum of particles  //
  /////////////////////////////////////////////////////////////////////////////
  else if (xdistribution_type == "mcutpTlep100" || xdistribution_type == "McutpTlep100"){
    int cut_condition = 0;
    if (osi->esi->event_particle_group[0][osi->RA_x_a][osi->esi->access_object_selection["nua"]].size() > 0){
      if (osi->esi->event_particle_group[0][osi->RA_x_a][osi->esi->access_object_selection["missing"]][0]->pTXX() < 100.){cut_condition = 1;}
    }
    if (!cut_condition){
      for (int i_p = 0; i_p < osi->esi->event_particle_group[0][osi->RA_x_a][osi->esi->access_object_selection["lep"]].size(); i_p++){
	if (osi->esi->event_particle_group[0][osi->RA_x_a][osi->esi->access_object_selection["lep"]][i_p]->pTXX() < 100.){cut_condition = 1; break;}
      }
    }

    if (cut_condition){
      observable_max = -1.;
    }
    else {
      for (int i_g = 0; i_g < particles.size(); i_g++) {
	observable_max += particle_pointer[i_g]->mXX();
      }
    }
  }



  ///////////////////////////////////////
  //  transverse-energy distributions  //
  ///////////////////////////////////////


  ///////////////////////////////////////////////////////////
  //  distribution in the (scalar) sum of ET of particles  //
  ///////////////////////////////////////////////////////////
  else if (xdistribution_type == "ET" || xdistribution_type == "HT") {
    double ET_scalar_sum=0;
    for (int group=0; group<particles.size(); group++) {
      ET_scalar_sum += particle_pointer[group]->ETXX();
    }
    observable = ET_scalar_sum;
  }
  /*
  // SK: Should be identical to ET definition:
  ///////////////////////////////////////////////////////////////////////////////
  //  transverse-mass distribution of particle group (sum when several groups) //
  //  HT = sum_i mT(particle i);  mT(particle i) = sqrt(m^2+pT^2)              //
  ///////////////////////////////////////////////////////////////////////////////
  else if (xdistribution_type == "HT") {
    double m_i, pT_i, mT_i;
    observable = 0.;
    // loop over all particle groups (group=0,1,2,..., 'particle group+1' distribution.dat)
    for (int group=0; group<particles.size(); group++) {
      m_i  = particle_pointer[group]->mXX();
      pT_i = particle_pointer[group]->pTXX();
      mT_i = sqrt(m_i*m_i+pT_i*pT_i);
      observable += mT_i; // scalar sum of mT of all particle groups
    }
  }
  */


  /////////////////////////////////////
  //  transverse-mass distributions  //
  /////////////////////////////////////


  ///////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //  transverse-mass distribution of one, two or more particle groups                                         //
  //  one particle group (theoretical definition):                                                             //
  //  mT(particle 1) = sqrt(m^2+pT^2)                                                                          //
  //    -> identical to ET                                                                                     //
  //  two particle groups (experimental definition, used when neutrinos plus one other particle involved):     //
  //  mT = sqrt{[ET(particle 1) + ET(particle 2)]^2 - [pTvec(particle 1)+pTvec(particle 2)]^2};                //
  //  ET(particle i) = sqrt(m_i^2+ pT_i^2)                                                                     //
  //  more than two particle groups (more than one non-neutrino involved):                                     //
  //  mT = sqrt{[ET(particle 1) + sum_i>1 ET(particle i)]^2 - [pTvec(particle 1)+sum_i pTvec(particle i)]^2};  //
  ///////////////////////////////////////////////////////////////////////////////////////////////////////////////
  else if (xdistribution_type == "mT" || xdistribution_type == "MT") {
    // use theoretical definition, when only one particle group
    if (particles.size() == 1){
      observable = particle_pointer[0]->ETXX(); // theoretical definition of tranverse mass
    }
    // use experimental definition, when two particle group
    else if (particles.size() == 2){
      // all neutrinos should go in the first reconstructed particle
      // experimental definition of tranverse mass when neutrinos and one other particle are involved
      //      observable = sqrt(pow(particle_pointer[0]->ETXX() + particle_pointer[1]->ETXX(), 2) - (particle_pointer[0] + particle_pointer[1])->pT2XX());
      // different definition in MATRIX: !!!
      observable = sqrt(pow(particle_pointer[0]->pTXX() + particle_pointer[1]->pTXX(), 2) - (*particle_pointer[0] + *particle_pointer[1]).pT2XX());
    }
    else if (particles.size() > 2){
      // all neutrinos should go into the first reconstructed particle
      // experimental definition of tranverse mass when neutrinos and more than one non-neutrino are involved
      particle sum_momentum;
      double sum_ET = 0.;
      for (int i_g = 0; i_g < particles.size(); i_g++) {
	//	sum_ET += particle_pointer[i_g]->ETXX();
	// different definition in MATRIX: !!!
	sum_ET += particle_pointer[i_g]->pTXX();
	sum_momentum = sum_momentum + *particle_pointer[i_g];
      }
      observable = sqrt(pow(sum_ET, 2) - sum_momentum.pT2XX());
    }
    /*
    // alternative definition for 2 involved massless particles (E == ET == |pT|):
    double dphi=abs(particle_pointer[0]->phiXX()*particle_pointer[1]->phiXX());
    observable = sqrt(2*particle_pointer[0].x0()*particle_pointer[1].x0()*(1.0-cos(dphi)));
    */
  }
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////
  //  transverse-mass distribution of particle/sum of particles                                               //
  //  mTCMS = sqrt(2 pT(particle 1) pT(sum_i>1 particle i) [1-cos(deltaPhi(particle1, sum_i>1 particle i))]   //
  //  for two particles same as mT, but only valid for massless particles; for more than two particles pT of  //
  //  vectorial sum instead of scalar sum of pT's (probably wrong!? FIXME)                                    //
  //////////////////////////////////////////////////////////////////////////////////////////////////////////////
  else if (xdistribution_type == "mTCMS" || xdistribution_type == "MTCMS") {
    assert(particles.size()>1  && "ERROR in xdistribution_type mTCMS: at least two particle groups required");

    // all neutrinos should go in the first reconstructed particle

    // sum momenta of all particle group after the first one
    particle temp_fvl;
    for (int group=1; group<particles.size(); group++) {
      temp_fvl = temp_fvl + *particle_pointer[group];
    }
    // compute difference in phi between (sum of) neutrino(s) and (sum of) other particle(s)
    double temp_dphi = abs(temp_fvl.phiXX() - particle_pointer[0]->phiXX());
    if (temp_dphi > pi){temp_dphi = 2 * pi - temp_dphi;}

    // compute transverse mass as defined by experiments when neutrinos involved for massless particles
    observable = sqrt(2 * temp_fvl.pTXX() * particle_pointer[0]->pTXX() * (1. - cos(temp_dphi)));
    // double observable2 = sqrt(pow(particle_pointer[0]->ETXX() + particle_pointer[1]->ETXX(), 2) - (particle_pointer[0] + particle_pointer[1])->pT2XX());
  }
  //
  // MW: pretty sure this is wrong (pT of vectorial sum, instead of scalar sum of pT's)
  // ==> THIS SHOULD NOT BE USED ANYMORE, mT should not include the correct formula
  //
  else if (xdistribution_type == "mTATLAS" || xdistribution_type == "MTATLAS") {
    //    particle temp_fvm = particle_pointer[0];
    //    double temp_pTm = temp_fvm->pTXX();

    particle temp_fvl;
    for (int group=1; group<particles.size(); group++) {
      temp_fvl = temp_fvl + *particle_pointer[group];
    }
    observable = sqrt(pow(temp_fvl.ETXX() + particle_pointer[0]->pTXX(), 2) - (temp_fvl + *particle_pointer[0]).pT2XX());
  }
  ////////////////////////////////////////////////////////////////////////
  //  transverse cluster mass distribution of paticle/sum of particles  //
  ////////////////////////////////////////////////////////////////////////
  //
  // MW: what is this needed for??? --> remove? (FIXME)
  //
  else if (xdistribution_type == "mTcluster" || xdistribution_type == "MTcluster") {
    assert(particles.size()==2);
    //    assert(particles.size()<=2);
    //    if (particles.size()==1) {
    // note: in contrast to old implementation, we *do* sum over the particle momenta
    //      observable = particle_pointer[0].mT();
    //    } else {
    observable = sqrt(pow(sqrt(particle_pointer[0]->m2XX() + particle_pointer[0]->pT2XX()) + particle_pointer[1]->pTXX(), 2) - (*particle_pointer[0] + *particle_pointer[1]).pT2XX());
    //    }
  }
  else if (xdistribution_type == "mTZZ") {
    assert(particles.size()==2);
    //    assert(particles.size()<=2);
    //    if (particles.size()==1) {
    // note: in contrast to old implementation, we *do* sum over the particle momenta
    //      observable = particle_pointer[0].mT();
    //    } else {
    double mZ = osi->msi->M_pdg[23];
    observable = sqrt(pow(sqrt(particle_pointer[0]->pT2XX()+mZ*mZ) + sqrt(particle_pointer[1]->pT2XX()+mZ*mZ), 2) - (*particle_pointer[0] + *particle_pointer[1]).pT2XX());
    //    }
  }


  ////////////////////////////////////
  //  invariant-mass distributions  //
  ////////////////////////////////////


  /////////////////////////////////////////////////
  //  invariant-mass 'distribution' of particle  //
  /////////////////////////////////////////////////
  else if (xdistribution_type == "m" || xdistribution_type == "M") {
    assert(particles.size() == 1 && "ERROR in xdistribution_type 'm/M': exactly one particle required");
    observable = particle_pointer[0]->mXX();
  }
  //////////////////////////////////////////////////////////////////////////
  //  invariant-mass-difference distribution of two particles: m_1 - m_2  //
  //  m(particle i) = sqrt(p^2)                                           //
  //////////////////////////////////////////////////////////////////////////
  else if (xdistribution_type == "dm" || xdistribution_type == "dM") {
    assert(particles.size() == 2 && "ERROR in xdistribution_type 'dm/dM': exactly two particles required");
    observable = particle_pointer[0]->mXX() - particle_pointer[1]->mXX();
  }
  //////////////////////////////////////////////////////////////////////////////
  //  |invariant-mass-difference| distribution of two particles: |m_1 - m_2|  //
  //////////////////////////////////////////////////////////////////////////////
  else if (xdistribution_type == "absdm" || xdistribution_type == "absdM") {
    assert(particles.size() == 2 && "ERROR in xdistribution_type 'absdm/absdM': exactly two particles required");
    observable = abs(particle_pointer[0]->mXX() - particle_pointer[1]->mXX());
  }
  ////////////////////////////////////////////////////////////
  //  maximum invariant-mass distribution of two particles  //
  ////////////////////////////////////////////////////////////
  else if (xdistribution_type == "mmax" || xdistribution_type == "Mmax") {
    // could be extended to >2 particles
    assert(particles.size() == 2 && "ERROR in xdistribution_type 'mmax/Mmax': exactly two particles required");
    observable = max(particle_pointer[0]->mXX(),particle_pointer[1]->mXX());
  }
  ////////////////////////////////////////////////////////////
  //  minimum invariant-mass distribution of two particles  //
  ////////////////////////////////////////////////////////////
  else if (xdistribution_type == "mmin" || xdistribution_type == "Mmin") {
    // could be extended to >2 particles
    assert(particles.size() == 2 && "ERROR in xdistribution_type 'mmin/Mmin': exactly two particles required");
    observable = min(particle_pointer[0]->mXX(),particle_pointer[1]->mXX());
  }
  /////////////////////////////////////////////////////////////////////////////
  //  cumulative invariant-mass 'distribution' of particle/sum of particles  //
  /////////////////////////////////////////////////////////////////////////////
  else if (xdistribution_type == "mcut" || xdistribution_type == "Mcut"){
    for (int i_g = 0; i_g < particles.size(); i_g++) {
      observable_max += particle_pointer[i_g]->mXX();
    }
  }
  ////////////////////////////////////////////////////////////////////////////////////////
  //  cumulative |invariant-mass-difference| distribution of particle/sum of particles  //
  ////////////////////////////////////////////////////////////////////////////////////////
  else if (xdistribution_type == "absdmcut" || xdistribution_type == "absdMcut"){
    assert(particles.size() == 2 && "ERROR in xdistribution_type 'absdmcut/absdMcut': exactly two particles required");
    observable_max = abs(particle_pointer[0]->mXX() - particle_pointer[1]->mXX());
  }


  ///////////////////////////////////////
  //  (pseudo-)rapidity distributions  //
  ///////////////////////////////////////


  ////////////////////////////////////////////
  //  rapidity distribution of particle: y  //
  ////////////////////////////////////////////
  else if (xdistribution_type == "y") {
    assert(particles.size() == 1 && "ERROR in xdistribution_type 'y': exactly one particle required");
    observable = particle_pointer[0]->rapidityXX();
  }
  //////////////////////////////////////////////////////////////////////////
  //  pseudo-rapidity distribution of particle: eta (= artanh(p_x3/|p|))  //
  //////////////////////////////////////////////////////////////////////////
  else if (xdistribution_type == "eta") {
    assert(particles.size() == 1 && "ERROR in xdistribution_type 'eta': exactly one particle required");
    observable = particle_pointer[0]->etaXX();
  }
  ////////////////////////////////////////////////
  //  |rapidity| distribution of particle: |y|  //
  ////////////////////////////////////////////////
  else if (xdistribution_type == "absy") {
    assert(particles.size() == 1 && "ERROR in xdistribution_type 'absy': exactly one particle required");
    observable = abs(particle_pointer[0]->rapidityXX());
  }
  /////////////////////////////////////////////////////////
  //  |pseudo-rapidity| distribution of particle: |eta|  //
  /////////////////////////////////////////////////////////
  else if (xdistribution_type == "abseta") {
    assert(particles.size() == 1 && "ERROR in xdistribution_type 'abseta': exactly one particle required");
    observable = abs(particle_pointer[0]->etaXX());
  }
  //////////////////////////////////////////////////////////////////////
  //  rapidity-difference distribution of two particles: (y_1 - y_2)  //
  //////////////////////////////////////////////////////////////////////
  else if (xdistribution_type == "dy") {
    assert(particles.size() == 2 && "ERROR in xdistribution_type 'dy': exactly two particles required");
    observable = particle_pointer[0]->rapidityXX() - particle_pointer[1]->rapidityXX();
  }
  ///////////////////////////////////////////////////////////////
  //  rapidity-sum distribution of two particles: (y_1 + y_2)  //
  ///////////////////////////////////////////////////////////////
  else if (xdistribution_type == "ysum") {
    assert(particles.size() == 2 && "ERROR in xdistribution_type 'ysum': exactly two particles required");
    observable = particle_pointer[0]->rapidityXX() + particle_pointer[1]->rapidityXX();
  }
  /////////////////////////////////////////////////////////////////////////////////
  //  pseudo-rapidity-difference distribution of two particles: (eta_1 - eta_2)  //
  /////////////////////////////////////////////////////////////////////////////////
  else if (xdistribution_type == "deta") {
    assert(particles.size() == 2 && "ERROR in xdistribution_type 'deta': exactly two particles required");
    observable = particle_pointer[0]->etaXX() - particle_pointer[1]->etaXX();
  }
  else if (xdistribution_type == "dabsy") {
    ////////////////////////////////////////////////////////////////////////
  //  |rapidity|-difference distribution of two particles: |y_1| - |y_2|  //
    ////////////////////////////////////////////////////////////////////////
    assert(particles.size() == 2 && "ERROR in xdistribution_type 'dabsy': exactly two particles required");
    observable = abs(particle_pointer[0]->rapidityXX()) - abs(particle_pointer[1]->rapidityXX());
  }
  /////////////////////////////////////////////////////////////////////////////////////
  //  |pseudo-rapidity|-difference distribution of two particles: |eta_1| - |eta_2|  //
  /////////////////////////////////////////////////////////////////////////////////////
  else if (xdistribution_type == "dabseta") {
    assert(particles.size() == 2 && "ERROR in xdistribution_type 'dabseta': exactly two particles required");
    observable = abs(particle_pointer[0]->etaXX()) - abs(particle_pointer[1]->etaXX());
  }
  ////////////////////////////////////////////////////////////////////////
  //  |rapidity-difference| distribution of two particles: |y_1 - y_2|  //
  ////////////////////////////////////////////////////////////////////////
  else if (xdistribution_type == "absdy") {
    assert(particles.size() == 2 && "ERROR in xdistribution_type 'absdy': exactly two particles required");
    observable = abs(particle_pointer[0]->rapidityXX() - particle_pointer[1]->rapidityXX());
  }
  /////////////////////////////////////////////////////////////////
  //  |rapidity-sum| distribution of two particles: |y_1 + y_2|  //
  /////////////////////////////////////////////////////////////////
  else if (xdistribution_type == "absysum") {
    assert(particles.size() == 2 && "ERROR in xdistribution_type 'absysum': exactly two particles required");
    observable = abs(particle_pointer[0]->rapidityXX() + particle_pointer[1]->rapidityXX());
  }
  ///////////////////////////////////////////////////////////////////////////////////
  //  |pseudo-rapidity-difference| distribution of two particles: |eta_1 - eta_2|  //
  ///////////////////////////////////////////////////////////////////////////////////
  else if (xdistribution_type == "absdeta") {
    assert(particles.size() == 2 && "ERROR in xdistribution_type 'absdeta': exactly two particles required");
    observable = abs(particle_pointer[0]->etaXX() - particle_pointer[1]->etaXX());
  }
  else if (xdistribution_type == "absdabsy") {
    //////////////////////////////////////////////////////////////////////////////
    //  ||rapidity|-difference| distribution of two particles: ||y_1| - |y_2||  //
    //////////////////////////////////////////////////////////////////////////////
    assert(particles.size() == 2 && "ERROR in xdistribution_type 'absdabsy': exactly two particles required");
    observable = abs(abs(particle_pointer[0]->rapidityXX()) - abs(particle_pointer[1]->rapidityXX()));
  }
  else if (xdistribution_type == "absdabseta") {
    /////////////////////////////////////////////////////////////////////////////////////////
    //  ||pseudo-rapidity|-difference| distribution of two particles: ||eta_1| - |eta_2||  //
    /////////////////////////////////////////////////////////////////////////////////////////
    assert(particles.size() == 2 && "ERROR in xdistribution_type 'dabsdeta': exactly two particles required");
    observable = abs(abs(particle_pointer[0]->etaXX()) - abs(particle_pointer[1]->etaXX()));
  }
  ///////////////////////////////////////////////////
  //  maximum-|rapidity| distribution of particle  //
  ///////////////////////////////////////////////////
  else if (xdistribution_type == "absymax") {
    // FIXME: we might want to change this such that it computes the maximum eta of the *reconstructed* particles
    assert(particles.size() == 1 && "ERROR in xdistribution_type 'absymax': exactly one particle required");
    double max_y=0;
    for (int particle=0; particle<particles[0].size(); particle++) {
      double tmp_y=abs(particles[0][particle].momentum.rapidity());
      if (tmp_y>max_y) {
        max_y=tmp_y;
      }
    }
    observable = max_y;
  }
  //////////////////////////////////////////////////////////
  //  maximum-|pseudo-rapidity| distribution of particle  //
  //////////////////////////////////////////////////////////
  else if (xdistribution_type == "absetamax") {
    // different definition in MATRIX !!!
    // FIXME: we might want to change this such that it computes the maximum eta of the *reconstructed* particles
    /*
    assert(particles.size() == 1 && "ERROR in xdistribution_type 'absetamax': exactly one particle required");
    double max_eta=0;
    for (int particle=0; particle<particles[0].size(); particle++) {
      double tmp_eta=abs(particles[0][particle].momentum.eta());
      if (tmp_eta>max_eta) {
        max_eta=tmp_eta;
      }
    }
    observable = max_eta;
    */
    observable = 0.;
    for (int group = 0; group < particles.size(); group++){
      if (observable < abs(particle_pointer[group]->etaXX())){observable = abs(particle_pointer[group]->etaXX());}
    }

  }
  ///////////////////////////////////////////////////
  //  minimum-|rapidity| distribution of particle  //
  ///////////////////////////////////////////////////
  else if (xdistribution_type == "absymin") {
    // FIXME: we might want to change this such that it computes the maximum eta of the *reconstructed* particles
    assert(particles.size() == 1 && "ERROR in xdistribution_type 'absymin': exactly one particle required");

    double min_y=1.e99;
    for (int particle=0; particle<particles[0].size(); particle++) {
      double tmp_y=abs(particles[0][particle].momentum.rapidity());
      if (tmp_y<min_y) {
        min_y=tmp_y;
      }
    }
    observable = min_y;
  }
  //////////////////////////////////////////////////////////
  //  minimum-|pseudo-rapidity| distribution of particle  //
  //////////////////////////////////////////////////////////
  else if (xdistribution_type == "absetamin") {
    // FIXME: we might want to change this such that it computes the maximum eta of the *reconstructed* particles
    assert(particles.size() == 1 && "ERROR in xdistribution_type 'absetamin': exactly one particle required");
    double min_eta=1.e99;
    for (int particle=0; particle<particles[0].size(); particle++) {
      double tmp_eta=abs(particles[0][particle].momentum.eta());
      if (tmp_eta<min_eta) {
        min_eta=tmp_eta;
      }
    }
    observable = min_eta;
  }


  ///////////////////////////////////////////
  //  angular (and similar) distributions  //
  ///////////////////////////////////////////


  ///////////////////////////////////////////////////////////////////
  //  transverse-angle distribution of particle a                  //
  //  if particle_2nd != 0, transverse-angle difference of a & b   //
  ///////////////////////////////////////////////////////////////////
  else if (xdistribution_type == "phi") {
    assert((particles.size() == 1 || particles.size() == 2) && "ERROR in xdistribution_type 'phi': exactly one or two particles required");
    if (particles.size() == 1) {
      observable = particle_pointer[0]->phiXX();
    } else {
      double dphi_abs = abs(particle_pointer[0]->phiXX()-particle_pointer[1]->phiXX());
      observable = min(f2pi-dphi_abs,dphi_abs);
    }
  }
  
  /////////////////////////////////////////////////////
  //  azimuthal-angle difference of particles a & b  //
  /////////////////////////////////////////////////////
  else if (xdistribution_type == "dphi") {
    assert((particles.size() == 2) && "ERROR in xdistribution_type 'dphi': exactly two particles required");
    double dphi_abs = abs(particle_pointer[0]->phiXX() - particle_pointer[1]->phiXX());
    observable = min(f2pi - dphi_abs, dphi_abs);
  }
  
  ////////////////////////////////////////////////////////////////
  //  normalized azimuthal-angle difference of particles a & b  //
  ////////////////////////////////////////////////////////////////
  else if (xdistribution_type == "dphi_pi") {
    assert((particles.size() == 2) && "ERROR in xdistribution_type 'dphi_pi': exactly two particles required");
    double dphi_abs = abs(particle_pointer[0]->phiXX() - particle_pointer[1]->phiXX());
    observable = min(f2pi - dphi_abs, dphi_abs) / pi;
  }
  
  /////////////////////////////////////////////////
  //  costheta distribution of particle          //
  //  if particle_2nd != 0, angle between 1 & 2  //
  /////////////////////////////////////////////////
  else if (xdistribution_type == "costheta") {
    assert((particles.size() == 1 || particles.size() == 2) && "ERROR in xdistribution_type 'costheta': exactly one or two particles required");
    if (particles.size()==1) {
      observable = (particle_pointer[0]->momentum).costheta();
    } else {
      observable = cosangle(particle_pointer[0]->momentum, particle_pointer[1]->momentum);
    }
  }
  ///////////////////////////////////////////////////////////////////
  //  transverse-angle--rapidity plane distribution of particle a  //
  ///////////////////////////////////////////////////////////////////
  else if (xdistribution_type == "dR") {
    assert(particles.size() == 2 && "ERROR in xdistribution_type 'dR': exactly two particles required");
    double temp_dphi = abs(particle_pointer[0]->phiXX() - particle_pointer[1]->phiXX());
    if (temp_dphi > pi){temp_dphi = 2 * pi - temp_dphi;}
    double dy = particle_pointer[0]->rapidityXX() - particle_pointer[1]->rapidityXX();
    observable = sqrt(pow(dy, 2) + pow(temp_dphi, 2));
  }
  //////////////////////////////////////////////////////////////////////////
  //  transverse-angle--pseudo-rapidity plane distribution of particle a  //
  //////////////////////////////////////////////////////////////////////////
  else if (xdistribution_type == "dReta") {
    assert(particles.size() == 2 && "ERROR in xdistribution_type 'dReta': exactly two particles required");
    double temp_dphi = abs(particle_pointer[0]->phiXX() - particle_pointer[1]->phiXX());
    if (temp_dphi > pi){temp_dphi = 2 * pi - temp_dphi;}
    double deta = particle_pointer[0]->etaXX() - particle_pointer[1]->etaXX();
    observable = sqrt(pow(deta, 2) + pow(temp_dphi, 2));
  }
  ///////////////////////////////////////////////////////////////////
  //      costhetastar variable for diphoton and Drell-Yan         //
  //       definition as in Collins and Soper, PRD 16 (1977) 2219  //
  ///////////////////////////////////////////////////////////////////
  else if (xdistribution_type == "abscosthetastar") {
    assert(particles.size() == 2 && "ERROR in xdistribution_type 'costhetastar': exactly two particles required");
    double eta_p1 = particle_pointer[0]->etaXX();
    double eta_p2 = particle_pointer[1]->etaXX();
    observable = abs(tanh((eta_p1 - eta_p2) / 2));
  }
  else if (xdistribution_type == "costhetastar") {
    assert(particles.size() == 2 && "ERROR in xdistribution_type 'costhetastar': exactly two particles required");
    // previous definition - not identical to abscosthetastar above !!!
    double pt3 = particle_pointer[0]->pTXX();
    double pt4 = particle_pointer[1]->pTXX();
    double eta3 = particle_pointer[0]->etaXX();
    double eta4 = particle_pointer[1]->etaXX();
    particle p34 = *particle_pointer[0] + *particle_pointer[1];
    double m34 = p34.mXX();
    double pt34 = p34.pTXX();
    observable = abs(2 * pt3 * pt4 * sinh(eta3 - eta4) / m34 / sqrt(pow(m34, 2) + pow(pt34, 2)));
  }

  ///////////////////////////////////////////////////////////////////
  //      phistar variable for Drell-Yan                           //
  //      definition as in arXiv; 1606.02330                       //
  ///////////////////////////////////////////////////////////////////
  else if (xdistribution_type == "phistar") {
    assert(particles.size() == 2 && "ERROR in xdistribution_type 'phistar': exactly two particles required");
    double phim = particle_pointer[0]->phiXX();
    double phip = particle_pointer[1]->phiXX();
    double etam = particle_pointer[0]->etaXX();
    double etap = particle_pointer[1]->etaXX();

    // compute difference in phi between l- and l+
    double dphi = abs(phim - phip);
    if (dphi > pi){dphi = 2 * pi - dphi;}
    double costhetastar = tanh((etam - etap) / 2);
    observable = tan( (pi-dphi)/2.)*sqrt(1.-pow(costhetastar,2)); 
  }
  //////////////////////////////
  //      costhetaS variable  //
  //////////////////////////////
  else if (xdistribution_type == "abscosthetaS") {
    assert(particles.size() == 2 && "ERROR in xdistribution_type 'costhetaS': exactly two particles required");
    double y_p1 = particle_pointer[0]->rapidityXX();
    double y_p2 = particle_pointer[1]->rapidityXX();
    observable = tanh(abs((y_p1 - y_p2) / 2));
  }

  ///////////////////////////////////////////////////////////////////////////////////////
  //  special distributions, which should become obsolete with user-defined particles  //
  ///////////////////////////////////////////////////////////////////////////////////////


  ////////////////////////////////////////////////////////////////////////////////
  //  delta phi of reconstructed Z's according to the CMS pairing prescription  //
  //  FIXME: should be replaced by a user defined distribution                  //
  ////////////////////////////////////////////////////////////////////////////////
  else if (xdistribution_type == "phi_pair_CMS") {
    assert(particles.size()==2);
    assert(particles[0].size()==2);
    assert(particles[1].size()==2);

    double dMZ1 = abs((particles[0][0].momentum+particles[1][0].momentum).m()-osi->msi->M_pdg[23]);
    double dMZ2 = abs((particles[0][0].momentum+particles[1][1].momentum).m()-osi->msi->M_pdg[23]);
    double dMZ3 = abs((particles[0][1].momentum+particles[1][0].momentum).m()-osi->msi->M_pdg[23]);
    double dMZ4 = abs((particles[0][1].momentum+particles[1][1].momentum).m()-osi->msi->M_pdg[23]);

    fourvector Z1,Z2;

    double dMZmin = min(dMZ1,min(dMZ2,min(dMZ3,dMZ4)));

    if (dMZ1 == dMZmin) {
      Z1 = particles[0][0].momentum+particles[1][0].momentum;
      Z2 = particles[0][1].momentum+particles[1][1].momentum;
    } else if (dMZ2 == dMZmin) {
      Z1 = particles[0][0].momentum+particles[1][1].momentum;
      Z2 = particles[0][1].momentum+particles[1][0].momentum;
    } else if (dMZ3 == dMZmin) {
      Z1 = particles[0][1].momentum+particles[1][0].momentum;
      Z2 = particles[0][0].momentum+particles[1][1].momentum;
    } else {
      assert(dMZ4==dMZmin);
      Z1 = particles[0][1].momentum+particles[1][1].momentum;
      Z2 = particles[0][0].momentum+particles[1][0].momentum;
    }

    observable = min(f2pi-abs(Z1.phi()-Z2.phi()),abs(Z1.phi()-Z2.phi()));
  }
  ///////////////////////////////////////////////////////////////////
  //  invariant mass of the subleading Z, according to the CMS pairing prescription //
  //  FIXME: should be replaced by a user defined distribution   //
  ///////////////////////////////////////////////////////////////////
  else if (xdistribution_type == "m34_CMS") {

    // each particle contains a list of Z boson candidates
    double dMZ_min = 1e99;
    int Z_group=-1;
    int first_Z=-1;

    for (int group=0; group<particles.size(); group++) {
      for (int particle=0; particle<particles[group].size()/2; particle++) {
        double dMZ = abs((particles[group][2*particle].momentum+particles[group][2*particle+1].momentum).m()-osi->msi->M_pdg[23]);
        if (dMZ < dMZ_min) {
          Z_group = group;
          first_Z = particle;
          dMZ_min = dMZ;
        }
      }
    }
    fourvector Z2 = particles[Z_group][(2*first_Z+2)%4].momentum+particles[Z_group][(2*first_Z+1+2)%4].momentum;

    observable = Z2.m();
  }
  /////////////////////////////////////////////////////////////////////
  //  costheta of the leading Z, evaluated in the rest frame of the  //
  //  four leptons, according to the CMS pairing prescription        //
  //  FIXME: should be replaced by a user defined distribution       //
  /////////////////////////////////////////////////////////////////////
  else if (xdistribution_type == "costheta12_CMS") {

    // each particle contains a list of Z boson candidates
    double dMZ_min = 1e99;
    int Z_group=-1;
    int first_Z=-1;

    for (int group=0; group<particles.size(); group++) {
      for (int particle=0; particle<particles[group].size()/2; particle++) {
        double dMZ = abs((particles[group][2*particle].momentum+particles[group][2*particle+1].momentum).m()-osi->msi->M_pdg[23]);
        if (dMZ < dMZ_min) {
          Z_group = group;
          first_Z = particle;
          dMZ_min = dMZ;
        }
      }
    }

    fourvector Z1 = particles[Z_group][2*first_Z].momentum+particles[Z_group][2*first_Z+1].momentum;
    fourvector Z2 = particles[Z_group][(2*first_Z+2)%4].momentum+particles[Z_group][(2*first_Z+1+2)%4].momentum;

    fourvector Z_boosted = Z1.boost(Z1+Z2);

    observable = abs(Z_boosted.costheta());
  }

  else {
    logger << LOG_FATAL << "distribution not implemented: " << xdistribution_type << endline_flush;
    exit(1);
  }
}

