// "The two-loop helicity amplitudes for gg -> V1V2 -> 4 leptons"
// A. von Manteuffel, L. Tancredi
//
// ggvvamp package


#include "ggvvamp.h"
#include <exception>
#include <sstream>
#include <ginac/ginac.h>
#include <cln/float.h>
#include <cln/rational_ring.h>

using namespace GiNaC;
using namespace std;

// general conversion functions

inline GiNaC::numeric torealnumeric(const GiNaC::ex& x) {
  const numeric delta_im(1e-5); // ignore unexpected imaginary parts smaller than this
  ex n = x.evalf();
  if (!GiNaC::is_a<numeric>(n))
      throw runtime_error("failed numerical evaluation (polylog ?)");
  if (abs(imag_part(x)) > delta_im)
    cerr << "ggvvamp: unexpected imaginary part (polylog ?)" << endl;
  return ex_to<numeric>(real_part(n));
}

// instantiation specific functions

#ifdef ARBITRARY_PRECISION

GiNaC::numeric tonumeric(const cln::cl_F& x) {
  return numeric(cln::cl_N(x));
}

GiNaC::numeric tonumeric(const cln::cl_N& x) {
  return numeric(x);
}

double todouble(const cln::cl_F& x) {
  return double_approx(x);
}

inline cln::cl_F toclf(const GiNaC::ex& x) {
  cln::cl_N n = torealnumeric(x).to_cl_N();
  if (cln::instanceof(n, cln::cl_RA_ring))
    return cln::cl_float(As(cln::cl_RA)(n), cln::default_float_format);
  //n = cln::cl_float(1.0, cln::default_float_format) * n;
  return cln::cl_float(As(cln::cl_F)(n), cln::default_float_format);
}

template<> cln::cl_F ggvvamp<cln::cl_F>::n(const GiNaC::ex& x) {
  return toclf(x);
}

std::ostream& operator<<(std::ostream& os, const cln::cl_F& x) {
  return os << tonumeric(x);
}

#endif

#ifdef QUAD_PRECISION

GiNaC::numeric tonumeric(const quadtype& x) {
  char buf[128];
  quadmath_snprintf(buf, sizeof buf, "%.35Qe", x);
  return numeric(buf);
}

double todouble(const quadtype& x) {
  return x;
}

inline quadtype toquad(const GiNaC::ex& x) {
  numeric n = torealnumeric(x);
  ostringstream ss;
  ss << n;
  quadtype q = strtoflt128(ss.str().c_str(), NULL);
  return q;
}

template<> quadtype ggvvamp<quadtype>::n(const GiNaC::ex& x) {
  return toquad(x);
}

std::ostream& operator<<(std::ostream& os, const quadtype& x) {
  return os << tonumeric(x);
}

#endif

#ifdef DOUBLE_PRECISION

GiNaC::numeric tonumeric(const double& x) {
  return numeric(x);
}

double todouble(const double& x) {
  return x;
}

inline double todouble(const GiNaC::ex& x) {
  numeric n = torealnumeric(x);
  return n.to_double();
}

template<> double ggvvamp<double>::n(const GiNaC::ex& x) {
  return todouble(x);
}

#endif

namespace {
void ggvvamp_welcome() {
  static bool first = true;
  if (first) {
    first = false;
    cout << "\n\"The two-loop helicity amplitudes for gg -> V1V2 -> 4 leptons\"\n";
    cout << "A. von Manteuffel, L. Tancredi\n";
    cout << "ggvvamp package, version 1.0\n" << endl;
  }
}
}

template<class T> ggvvamp<T>::ggvvamp(int Nf, IRscheme ir)
  : Nf(Nf), ir(ir), compute_gpls(true) {
  ggvvamp_welcome();
  r = new T[rsize];
  for (int j = 0 ; j <= nAcoeff; ++j) {
    for (int l = 0 ; l < nloop ; ++l) {
      for (int m = 0 ; m < nreim; ++m) {
        // only some coefficients will be computed, help to detect invalid usage
        A[j][l][m] = 1e100;
      }
    }
  }
}

template<class T> ggvvamp<T>::~ggvvamp() {
  delete[] r;
}

namespace {
template<unsigned int n> inline numeric Pow( const numeric& x ) { return x*Pow<n-1>(x); }
template<> inline numeric Pow<0>( const numeric& x ) { return numeric(1); }
}

// for arbitrary precision type enforce calculation in specified precision
// even if input was constructed from low precision numbers
template<class T> numeric adjustdigits(const T& x) {
  return tonumeric(x);
}
template<> numeric adjustdigits(const cln::cl_F& x) {
  return numeric(cln::cl_float(x, cln::default_float_format));
}

// convert from qt-subtraction to Catani's original subtraction scheme
template<class T>
void convert_A_to_Catani(ggvvamp<T>& c, int Nf) {
  T (*n)(const ex&) = &ggvvamp<T>::n;
  
  numeric CF = numeric(4,3);
  numeric CA = numeric(3);
  numeric TF = numeric(1,2);
  ex b0 = numeric(11,6)*CA - numeric(2,3)*TF*Nf;
    
  ex I1toCatani = -numeric(1,2)*Pi*Pi*CA + I*Pi*b0;

  I1toCatani = I1toCatani.evalf();
  
  for (int i = 1; i <= nAcoeff; ++i) {
    ex A1qt = tonumeric(c.A[i][0][0]) + I*tonumeric(c.A[i][0][1]);
    ex A2qt = tonumeric(c.A[i][1][0]) + I*tonumeric(c.A[i][1][1]);
    ex A1Catani = A1qt;
    ex A2Catani = A2qt + I1toCatani*A1qt;
    c.A[i][0][0] = n(real_part(A1Catani));
    c.A[i][0][1] = n(imag_part(A1Catani));
    c.A[i][1][0] = n(real_part(A2Catani));
    c.A[i][1][1] = n(imag_part(A2Catani));
  }
}

template<class T>
void ggvvamp<T>::compute(T s_n, T t_n, T ma2_n, T mb2_n) {

  // for convenience: prepare input with GiNaC::numeric
  numeric s = adjustdigits(s_n);
  numeric t = adjustdigits(t_n);
  numeric ma2 = adjustdigits(ma2_n);
  numeric mb2 = adjustdigits(mb2_n);
  numeric u = ma2 + mb2 - s - t;

  // check whether phase space point in physical region
  
  if (ma2 <= 0 || mb2 <= 0)
    throw runtime_error("illegal ma2 or mb2 value");
  
  numeric smin = Pow<2>(sqrt(ma2) + sqrt(mb2));
  if (s < smin)
    throw runtime_error("illegal s value");

  numeric root = sqrt(Pow<2>(s-ma2-mb2)-4*ma2*mb2); // kappa
  numeric t0 = (ma2 + mb2 - s - root)/2;
  numeric t1 = (ma2 + mb2 - s + root)/2;
  if (t < t0 || t > t1) {
    cerr << "t value should be within [" << t0 << ", " << t1 << "]" << endl;
    throw runtime_error("illegal t value");
  }

  // prepare input to generated code
  numeric x =  -t0 / ma2 ;
  numeric y =  t1*(u+t)/(ma2*mb2) - 1 ; 
  numeric z =  t1*t / (ma2*mb2) ;
 
  bool verbose = false;
  if (verbose) {
    cout << "s = " << s << "\t" << cln::float_precision(As(cln::cl_F)(s.to_cl_N())) << "\n";
    cout << "t = " << t << "\t" << cln::float_precision(As(cln::cl_F)(t.to_cl_N())) << "\n";
    cout << "u = " << u << "\t" << cln::float_precision(As(cln::cl_F)(u.to_cl_N())) << "\n";
    cout << "ma2 = " << ma2 << "\t" << cln::float_precision(As(cln::cl_F)(ma2.to_cl_N())) << "\n";
    cout << "mb2 = " << mb2 << "\t" << cln::float_precision(As(cln::cl_F)(mb2.to_cl_N())) << "\n";
    cout << "x = " << x << "\t" << cln::float_precision(As(cln::cl_F)(x.to_cl_N())) << "\n";
    cout << "y = " << y << "\t" << cln::float_precision(As(cln::cl_F)(y.to_cl_N())) << "\n";
    cout << "z = " << z << "\t" << cln::float_precision(As(cln::cl_F)(z.to_cl_N())) << endl;
  }
  
  p0001(n(ma2), n(t/ma2), n(u/ma2), n(mb2/ma2), n(x), n(y), n(z));
  
  switch (ir) {
    case IRschemeQt:
      break;
    case IRschemeCatani:
      convert_A_to_Catani(*this, Nf);
      break;
  }
}



#include "ggvvamp.xi"
