#pragma once

#include <FunctionID.h>
#include <Constants.h>

#include <unordered_map>
#include <complex>
#include <mutex>
#include <cassert>
#include <vector>
#include <memory>

#include "FivePointAmplitudes_config.h"

#ifdef FivePointAmplitudes_QD_ENABLED 
#include <qd/dd_real.h>
#endif // FivePointAmplitudes_QD_ENABLED

namespace FivePointAmplitudes {
using namespace PentagonFunctions;

class FunctionsManager {
  public:
    using R = double;
    using C = std::complex<double>;

#ifdef FivePointAmplitudes_QD_ENABLED 
    using RHP = dd_real;
    using CHP = std::complex<dd_real>;
#endif // FivePointAmplitudes_QD_ENABLED

  private:
    std::unordered_map<FunctionID, size_t> _function_indices;

    std::vector<FunctionObjectType<double>> _evaluators;

    std::vector<C> last_values;
    Kin<R> last_kin{3., -1., 1., 1., -1.};

#ifdef FivePointAmplitudes_QD_ENABLED 
    std::vector<FunctionObjectType<RHP>> _evaluators_HP;
    std::vector<CHP> last_values_HP;
    Kin<RHP> last_kin_HP{3., -1., 1., 1., -1.};
#endif // FivePointAmplitudes_QD_ENABLED


  public:
    /**
     * Add a function to the manager
     */
    size_t add(const FunctionID&);
    /**
     * Evaluate the functions and return the result.
     * Consecutive calls with the same kinematics evaluate only once.
     */
    const std::vector<C>& eval(const Kin<R>&);
#ifdef FivePointAmplitudes_QD_ENABLED
    const std::vector<CHP>& eval(const Kin<RHP>&);
    const std::vector<CHP>& eval_HP(const Kin<RHP>&);
#endif // FivePointAmplitudes_QD_ENABLED

    size_t size() const { return _evaluators.size(); }

  private:
    bool locked{false}; /// will not allow adding more functions once locked after the first evaluation
};

/**
 * Represents (abstractly) the type of entry in the basis of functions
 * f    - function
 * bc   - real boundary constant
 * bci  - imaginary boundary constant
 * str5 - sign(Im(tr5))
 * one  - literally one
 */
enum class f_type { f, bc, bci, str5, one };
f_type from_string(std::string);

} // namespace FivePointAmplitudes
