#pragma once
#include "../FunctionsManager.h"
#include "../RemaindersList.h"

#include "aux.hpp"

#include <exception>

namespace FivePointAmplitudes {
    
template <typename T> class qqyyy_h2 {
  private:
    std::shared_ptr<FunctionsManager> f_manager{new FunctionsManager};
    remainder_2q3y_1l_pmppp_12345<T> ppp_1l{};
    remainder_2q3y_1l_pmmpp_12345<T> mpp_1l{f_manager};
    //TODO: wrapping in shared_ptr is just a hack to not use copy/move-assignment operators for remainder classes, which are broken for some reason. 
    std::shared_ptr<remainder_2q3y_2l_pmmpp_12345<T>> mpp_2l;
    std::shared_ptr<remainder_2q3y_2l_pmmpp_nf_12345<T>> mpp_2l_nf;
    /**
     * The variable which selects which of 4 two-loop contributions to compute.
     * By default only h(2,0) is computed.
     */
    std::array<bool, 4> included_twoloop_contributions;

  public:
    using ReturnType = std::vector<std::vector<T>>;
    /**
     * Include only specified contributions
     * the list corresponds to {H(2,0),H(2,1),H(2,Nf),H(2,Nfy)} in [arXiv:2010.04681]
     */
    qqyyy_h2(const std::array<bool,4>& contributions = {true, false, false ,false});
    /**
     * Evaluate with s_ij invariants at scale musq.
     * If the scale is not given, it will be set to s_12.
     * 
     * @return {h_0, H_1, H_2}, H_1 and H_2 are dimensionless functions, normalized by h_0,
     * H_2 is {H(2,0),H(2,1),H(2,Nf),H(2,Nfy)} as defined in [arXiv:2010.04681]
     * Only planar contributions h^(2,0) and h(2,Nf) are computed for now. Others are simply set to zero.
     * The contributions DO NOT include any color factors in them.
     */
    ReturnType operator()(const std::array<T, 5>& v_in, T musq = 0);
};

template <typename T> qqyyy_h2<T>::qqyyy_h2(const std::array<bool, 4>& contributions) : included_twoloop_contributions(contributions) {
    if(contributions[0]) {
        mpp_2l = std::make_shared<remainder_2q3y_2l_pmmpp_12345<T>>(f_manager);
    }
    if(contributions[1]) {
        throw std::runtime_error("H(2,1) not implemented");
    }
    if(contributions[2]) {
        mpp_2l_nf = std::make_shared<remainder_2q3y_2l_pmmpp_nf_12345<T>>(f_manager);
    }
    if(contributions[3]) {
        throw std::runtime_error("H(2,Nfy) not implemented");
    }
}

template <typename T> typename qqyyy_h2<T>::ReturnType qqyyy_h2<T>::operator()(const std::array<T, 5>& v_in, T musq) {

    // set the scale to s_12 by default
    if (musq == 0) musq = v_in[0];

    ReturnType hi{{0},{0},{0,0,0,0}};

    sij5<T> sij(v_in);

    std::vector<std::array<int, 5>> permutations = {
        {1, 2, 3, 4, 5}, {1, 2, 4, 3, 5}, {1, 2, 5, 3, 4}, {2, 1, 3, 4, 5}, {2, 1, 4, 3, 5}, {2, 1, 5, 3, 4},
    };

    for (const auto& perm : permutations) { hi[0][0] += qp_qm_m_p_p_tree_squared(sij.permute(perm)); }
    // born ME2 is parity-invariant, so we don't need to sum over 12 helicities, but can multiply by 2
    hi[0][0] *= 2;

    // contributions with non-vanishing tree
    for (const auto& perm : permutations) {
        auto sijp = sij.permute(perm);
        auto treesq_h = qp_qm_m_p_p_tree_squared(sijp);
        // we normalize by the whole born matrix element already here to make all contributions close to 1
        treesq_h /= hi[0][0];
        hi[1][0] += treesq_h * auxiliary::eval_h1_h([&](const sij5<T>& sij, int p) { return mpp_1l.eval(sij, p, musq); }, sijp);

        // Nc^2/4 contribution
        if (included_twoloop_contributions[0]) {
            hi[2][0] += treesq_h * auxiliary::eval_h2_h([&](const sij5<T>& sij, int p) { return mpp_1l.eval(sij, p, musq); },
                                                        [&](const sij5<T>& sij, int p) { return mpp_2l->eval(sij, p, musq); }, sijp);
        }
        // Nf/Nc contribution
        if (included_twoloop_contributions[2]) { hi[2][2] += treesq_h * (mpp_2l_nf->eval(sijp, 1, musq).real() + mpp_2l_nf->eval(sijp, -1, musq).real()); }
    }

    // contributions from (1-loop)^2 with vanishing tree
    if (included_twoloop_contributions[0]) {
        for (const auto& perm : {permutations[0], permutations[3]}) {
            using auxiliary::abs2;
            auto sijp = sij.permute(perm);
            auto norm = qp_qm_p_p_p_norm_squared(sijp);
            norm /= hi[0][0];
            hi[2][0] += (abs2(ppp_1l.eval(sijp, 1, musq)) + abs2(ppp_1l.eval(sijp, -1, musq))) * norm;
        }
    }

    return hi;
}

extern template class qqyyy_h2<double>;
#ifdef FivePointAmplitudes_QD_ENABLED
extern template class qqyyy_h2<dd_real>;
#endif // FivePointAmplitudes_QD_ENABLED



} // FivePointAmplitudes
